function v = RB_2SLS(mu1, mu2, sigma, ZZ, N)
%RB_2SLS Rao-Blackwellized 2SLS estimation.
%   The line pi = W(:, numel(mu1) + 1 : end); assumes that pi is a vector,
%   that is, there is only one endogenous regressor.
%
%   mu1     GLS gamma (pi*beta)
%   mu2     GLS pi
%   sigma   Var(GLS gamma pi)
%   ZZ      Z'*Z
%   N       Number of draws for numerical expectation

    assert(isvector(mu1) && isvector(mu2) && ismatrix(sigma) && ismatrix(ZZ))
    assert(all(size(sigma) == numel(mu1) + numel(mu2)))
    assert(all(size(ZZ) == numel(mu2)))
    assert(mod(numel(mu2), numel(mu1)) == 0)

    if nargin < 5
        N = 100;
    else
        assert(isscalar(N) && N > 0 && fix(N) == N)
    end

    W = mvnrnd([mu1(:); mu2(:)], sigma, N);
    phi = W(:, 1 : numel(mu1));
    pi = W(:, numel(mu1) + 1 : end);
    piz = pi * ZZ;
    v = mean(sum(piz .* pi, 2) .\ sum(piz .* phi, 2));

end
