function v = RB_optIV(mu1, mu2, Var_OLS, Var_GLS, Var_GLS3, ind, UV_hat, Sigma_UV_hat, X, Y, Z, cZ, nZ, hZ, c2Z, N)
%RB_OPTIV Rao-Blackwellized Optimal IV estimation with first-stage included.
%   This is a modification of RB_optIV to include first-stage as a moment
%   condition.
%   mu1     Second-stage reduced-form GLS gamma. (k-by-1)
%   mu2     First-stage reduced-form GLS pi. (k-by-d)
%   (mu3)   GLS (pi; 0_d). ((1+d)k-by-d)
%   sigma   Variance-covariance of noise of (mu1, mu2, mu3).
%   X       Regressors.
%   Y       Dependent variables.
%   Z       Instrumental variables.
%   cZ      Categories of instrumental variables (integers 1, ..., nZ).
%   nZ      Number of categories.
%   N       Size of simulation. Default N = 200.
%           Does not iterate over estimation of structural error.
%           So N for the computation of reduced-form WLS.

%     assert(isvector(mu1) && isvector(mu2) && ismatrix(sigma))
%     assert(all(size(sigma) == numel(mu1) + numel(mu2)))
%     assert(mod(numel(mu2), numel(mu1)) == 0)

    if nargin < 16
        N = 50;
    else
        assert(isscalar(N) && N > 0 && fix(N) == N)
    end

    d = size(mu2, 2);
    k = size(mu2, 1);

    rho3 = zeros((1+d)*k, (1+d)*k); % Temp variable for Optimal IV without pi
    temp1 = zeros((1+d)+(1+d)*d);
    temp3 = zeros((1+d)*k*(1+d));

    mu3 = [mu2; zeros(d*k, d)];

    % Draw simulated reduced-form coefficients
    A = mvnrnd([mu1(:); mu2(:)], nearestSPD(Var_OLS-Var_GLS), N);
    % Extract simulated second-stage reduced-form coefficients
    phi = A(:, 1 : numel(mu1));
    % Extract simulated first-stage reduced-form coefficients
    pi = A(:, numel(mu1) + 1 : end);
    piz = pi * (Z' * Z);
    % 2SLS estimators of beta (N-by-1 vector of 2SLS)
    TSLS = sum(piz .* pi, 2) .\ sum(piz .* phi, 2);
    optIV3 = zeros(size(TSLS));

    % Prepare for the loop (get unique values of Z)
    Sigma_EV_hat = zeros(1+d, 1+d, nZ); % Covariance for eps (not consistent)

    d = size(mu2, 2);

    % Loop of computing optimal IV estimators
    for j = 1 : N
        % Estimate heteroskedasticity (inconsistent) and compute optimal IV
        EV_hat = [Y-X*TSLS(j) UV_hat(:, 2 : end)];
        for i = 1 : nZ
            Sigma_EV_hat(:, :, i) = E2(EV_hat(cZ==i, :));
        end

        rho3(:) = 0;
        temp3(:) = 0;
        for i = 1 : nZ
            optZtemp = blkdiag(c2Z(i,:)', c2Z(i,:)') / Sigma_EV_hat(:,:,i);
            rho3 = rho3 + hZ(i) * optZtemp*kron(eye(1+d), c2Z(i,:));
            optZtemp = kron(eye(1+d), optZtemp);
            temp1(ind>0, ind>0) = Sigma_UV_hat(i, ind(ind>0), ind(ind>0));
            temp3 = temp3 + hZ(i) * optZtemp*temp1*optZtemp';
        end

        Var_WLS3 = kron(eye(1+d), rho3) \ temp3 / kron(eye(1+d), rho3);
        optIV3(j) = RB_optIV3_inner(mu1, mu2, mu3, nearestSPD(Var_WLS3-Var_GLS3, true));
    end
    v = mean(optIV3);

    function v = RB_optIV3_inner(mu1, mu2, mu3, sigma)

        M = 100;

        % Draw simulated reduced-form coefficients
        AA = mvnrnd([mu1(:); mu2(:); mu3(:)]', sigma, M);
        c31a = AA(:, 1 : numel(mu1) + numel(mu2));
        c32a = reshape(permute(AA(:, numel(mu1) + numel(mu2) + 1 : end), [2 1]), [(1+d)*k d M]);
        c32 = zeros((1+d)*k, d*(1+k));
        c32(end-d*k+1 : end, end-d*k+1 : end) = eye(d*k);
        optTSLS = zeros(M, 1);
        for jj = 1 : M
            c32(:,1:d) = c32a(:,:,jj);
            temp = c32 \ c31a(jj, :)';
            optTSLS(jj) = temp(1);
        end
        v = mean(optTSLS);

    end

end
