%% Matlab simulation script for Kaji (2019) "Theory of Weak Identification in Semiparametric Models".

% Fix random seed for replication
seed = 1;
rng(seed)

% Weak asymptotics (isweak = true) or strong asymptotics (isweak = false)
isweak = true;

% Simulation design
nsim = 5000;             % Simulation size
n = 1000;               % Sample size
d = 1;                  % Number of endogenous regressors
k = 3;                  % Number of instruments
nZ = 2^k;               % Number of Z categories
hZ = zeros(nZ, 1);      % Histogram of Z categories
Z2c = 2.^(0 : k-1)';    % For conversion: Z -> category
c2Z = fliplr(dec2bin((1:nZ)-1)-'0')*2-1; % For conversion: category -> Z

% True parameters
beta = ones(d, 1);
if isweak
    pi = ones(k, d) / sqrt(n);
else
    pi = ones(k, d);
end
% Square-root of Var(U,V|Z)
temp = psdrnd(1+d, nZ);
rtSigmaUV = cell(nZ, 1);
for i = 1 : nZ      % Make it symmetric
    rtSigmaUV{i} = temp(:,:,i);
end

% Table 1: Heteroskedasticity Var(E,U,V|Z)
Sigma = zeros(2+d,2+d,nZ);
for i = 1 : nZ
    Sigma(:,:,i) = round([[1 -beta']; eye(1+d)]*rtSigmaUV{i}'*rtSigmaUV{i}*[[1 -beta']; eye(1+d)]', 2);
end

% Concentration parameter
cp = zeros(d);
for i = 1 : nZ
    temp = sqrtm(rtSigmaUV{i}'*rtSigmaUV{i});
    cp = cp + temp(2:end, 2:end) \ (pi'*c2Z(i,:)'*c2Z(i,:)*pi) / temp(2:end, 2:end);
end
cp = cp / nZ;

% Estimators
TSLS = zeros(d, nsim);      % 2SLS
RBTSLS = zeros(d, nsim);    % Rao-Blackwellized 2SLS
optIV = zeros(d, nsim);     % Optimal IV
RBoptIV = zeros(d, nsim);   % Rao-Blackwellized optimal IV

% Auxiliary estimators
SigmaUVhat = zeros(nZ, 1+d, 1+d); % Estimated Var(U,V|Z) (consistent)
SigmaEVhat = cell(nZ, 1); % Estimated Var(E,V|Z) (inconsistent)
optZ = zeros(k+d*k, 1+d, n);        % Optimal IV
VarWLS = zeros((1+d)*k*(1+d));     % Var(WLS) for optimal IV
VarGLS2 = zeros((1+d)*k*(1+d));    % Var(GLS) for optimal IV

% Temporary variables for computing optimal IV
temp1 = zeros((1+d)+(1+d)*d);
temp2 = zeros((1+d)*k*(1+d));
rho1 = zeros((1+d)*k, 1);
rho2 = zeros((1+d)*k, d*(1+k));
rho3 = zeros((1+d)*k, (1+d)*k);

% Utility variable for computing Var(GLS)
sp_sub = [kron(ones(1,1+d),1:n:d*n+1)' kron(1:n:d*n+1,ones(1,1+d))'];

% Utility variables for computing Var(WLS)
ind = [(1 : 1+d) zeros(1, (1+d)*d)];
for i = 1 : d
    ind((1+d)+1 + (1+d)*(i-1)) = i + 1;
end
ind2 = [(1 : (1+d)*k) zeros(1, (1+d)*k*d)];
for i = 1 : d
    ind2((1+d)*k+1 + (1+d)*k*(i-1) : (1+d)*k+1 + (1+d)*k*(i-1) + k-1) = (1 : k) + i * k;
end

% Initialize the progress bar
progbar(0)

for j = 1 : nsim

    %% Data

    % Instruments
    Z = randi([0 1], n, k);
    % Z category
    cZ = Z * Z2c + 1;
    % Normalize instruments to mean zero
    Z(Z==0) = -1;
    % Histogram of Z (for optIV3)
    hZ = histcounts(cZ, nZ);

    % Errors
    UV = normrnd(0, 1, n, 1+d);
    for i = 1 : nZ
        UV(cZ==i, :) = UV(cZ==i, :) * rtSigmaUV{i};
    end

    % Variables
    X = Z * pi + UV(:, 2:end);
    Y = Z * pi * beta + UV(:, 1);

    %% Reduced-form Estimators

    % Reduced-form feasible GLS
    UVhat = [Y X] - Z/(Z'*Z)*(Z'*[Y X]);
    for i = 1 : nZ
        SigmaUVhat(i,:,:) = E2(UVhat(cZ==i, :));
    end
    temp = SigmaUVhat(cZ, :);
    Omegahat = spdiags2(temp, sp_sub, (d+1)*n, (d+1)*n);
    Zbig = kron(eye(1+d), Z);
    GLS = (Zbig'/Omegahat*Zbig)\(Zbig'/Omegahat*[Y; X(:)]);
    GLS = reshape(GLS, k, 1+d);
    VarGLS = pinv(Zbig'/Omegahat*Zbig);
    UVZGLS = dotkron(UVhat, Z);
    VarOLS = (Zbig'*Zbig)\(UVZGLS'*UVZGLS)/(Zbig'*Zbig);

    %% Structural Estimators

    % 2SLS
    TSLS(j) = ((X'*Z)/(Z'*Z)*(Z'*X))\((X'*Z)/(Z'*Z)*(Z'*Y));

    % Rao-Blackwellized 2SLS
    RBTSLS(j) = RB_2SLS(GLS(:, 1), GLS(:, 2:end), nearestSPD(VarOLS-VarGLS, true), Z'*Z);

    % Optimal IV
    EVhat = [Y - X*TSLS(j) UVhat(:, 2:end)];
    for i = 1 : nZ
        SigmaEVhat{i} = E2(EVhat(cZ==i, :));
    end
    rho1(:) = 0;
    rho2(:) = 0;
    for i = 1 : n
        optZ(:,:,i) = blkdiag(Z(i,:)', Z(i,:)') / SigmaEVhat{cZ(i)};
        rho1 = rho1 + optZ(:,:,i)*[Y(i); X(i,:)'];
        rho2 = rho2 + optZ(:,:,i)*blkdiag(X(i,:), Z(i,:));
    end
    temp = rho2 \ rho1;
    optIV(j) = temp(1);

    % Rao-Blackwellized optimal IV
    rho3(:) = 0;
    temp2(:) = 0;
    for i = 1 : nZ
        optZtemp = blkdiag(c2Z(i,:)', c2Z(i,:)') / SigmaEVhat{i};
        rho3 = rho3 + hZ(i) * optZtemp*kron(eye(1+d), c2Z(i,:));
        temp1(ind>0, ind>0) = SigmaUVhat(i, ind(ind>0), ind(ind>0));
        temp2 = temp2 + hZ(i) * kron(eye(1+d), optZtemp)*temp1*kron(eye(1+d), optZtemp)';
    end
    % Variance of (c31, vec(c32)) (sort of)
    VarWLS = kron(eye(1+d), rho3) \ temp2 / kron(eye(1+d), rho3);
    VarGLS2(ind2>0, ind2>0) = VarGLS(ind2(ind2>0), ind2(ind2>0));
    RBoptIV(j) = RB_optIV(GLS(:,1), GLS(:,2:end), VarOLS, VarGLS, VarGLS2, ind, UVhat, SigmaUVhat, X, Y, Z, cZ, nZ, hZ, c2Z);

    % Progress bar
    progbar(j, nsim, 250)

end


%% Save variables

if isweak
    save('weak.mat', 'nsim', 'n', 'd', 'k', 'nZ', 'beta', 'pi', ...
        'Sigma', 'cp', 'TSLS', 'RBTSLS', 'optIV', 'RBoptIV')
    load('weak.mat')
else
    save('strong.mat', 'nsim', 'n', 'd', 'k', 'nZ', 'beta', 'pi', ...
        'Sigma', 'cp', 'TSLS', 'RBTSLS', 'optIV', 'RBoptIV')
    load('strong.mat')
end

% Convex loss table
T = table('Size', [2 4], ...
    'VariableTypes', {'double', 'double', 'double', 'double'}, ...
    'VariableNames', {'TSLS', 'RBTSLS', 'OptIV', 'RBOptIV'}, ...
    'RowNames', {'MSE', 'MAE'});

% Mean squared error
T{1,1} = mean((TSLS - beta).^2, 2);
T{1,2} = mean((RBTSLS - beta).^2, 2);
T{1,3} = mean((optIV - beta).^2, 2);
T{1,4} = mean((RBoptIV - beta).^2, 2);

% Mean absolute error
T{2,1} = mean(abs(TSLS - beta), 2);
T{2,2} = mean(abs(RBTSLS - beta), 2);
T{2,3} = mean(abs(optIV - beta), 2);
T{2,4} = mean(abs(RBoptIV - beta), 2);

% Table 2
round(T{:,:}, 3)


%% Figures for presentation

if isweak
    xl = beta + [-1.5 1.5];     % X Range
    yl = [0 1200];              % Y Range
    xe = xl(1) : 0.05 : xl(2);  % X Edges
else
    xl = beta + [-0.1 0.1];     % X Range
    yl = [0 1000];              % Y Range
    xe = xl(1) : 0.004 : xl(2); % X Edges
end


%% Figures for paper

% Figure 1/2 (a): 2SLS vs Rao-Blackwellized 2SLS
fig2(1, TSLS, RBTSLS, beta, sprintf('\\ 2SLS (MSE = %.2f)', T{1,1}), sprintf('\\ RB 2SLS (MSE = %.2f)', T{1,2}), xl, yl, xe)
if isweak
    print('weak2SLS.eps', '-depsc')
else
    print('strong2SLS.eps', '-depsc')
end

% Figure 1/2 (b): Optimal IV vs Rao-Blackwellized optimal IV
fig2(5, optIV, RBoptIV, beta, sprintf('\\ Optimal IV (MSE = %.2f)', T{1,3}), sprintf('\\ RB Optimal IV (MSE = %.2f)', T{1,4}), xl, yl, xe)
if isweak
    print('weakOptIV.eps', '-depsc')
else
    print('strongOptIV.eps', '-depsc')
end
