function B = nearestSPD(A, semi)
% nearestSPD - the nearest (in Frobenius norm) Symmetric Positive Definite matrix to A
% usage: Ahat = nearestSPD(A)
%
% From Higham: "The nearest symmetric positive semidefinite matrix in the
% Frobenius norm to an arbitrary real matrix A is shown to be (B + H)/2,
% where H is the symmetric polar factor of B=(A + A')/2."
%
% http://www.sciencedirect.com/science/article/pii/0024379588902236
%
% arguments: (input)
%  A - square matrix, which will be converted to the nearest Symmetric
%    Positive Definite Matrix.
%
% Arguments: (output)
%  B - The matrix chosen as the nearest SPD matrix to A.

    if nargin < 2
        semi = false;
    end

    % test for a square matrix A
    [r, c] = size(A);
    if r ~= c
        error('A must be a square matrix.')
    elseif r == 1
        B = real(A);
        % A was scalar and non-positive, so just return eps
        if B <= 0
            B = eps;
        end
        return
    end

    % symmetrize A into B
    if issymmetric(A)
        B = A;
    else
        B = (A + A') / 2;
    end

    % Compute the symmetric polar factor of B.
    [~, S, V] = svd(B);
    B = (B + V * S * V');
    B = (B + B') / 4;

    if semi
        mineig = min(eig(B));
        k = 0;
        while mineig < 0 % includes when mineig is complex
            k = k + 1;
            B = B + (abs(mineig) * k^2 + eps) * eye(size(A));
            mineig = abs(min(eig(B)));
        end
    else
        % test that Ahat is in fact PD. if it is not so, then tweak it just a bit.
        p = 1;
        k = 0;
        while p ~= 0
            [~, p] = chol(B);
            k = k + 1;
            if p ~= 0
                % Ahat failed the chol test. It must have been just a hair off,
                % due to floating point trash, so it is simplest now just to
                % tweak by adding a tiny multiple of an identity matrix.
                mineig = abs(min(eig(B)));
                B = B + (mineig * k^2 + eps) * eye(size(A));
            end
        end
    end

end
